import base64
import json
import os
from typing import Union

from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


class Encryption:
    """
    Encrypt & Decrypt data using AES-256-CBC
    """

    @staticmethod
    def generate_key() -> bytes:
        return os.urandom(32)

    @staticmethod
    def encrypt(data: Union[int, str, list, dict], key: bytes) -> bytes:
        # Convert the data to JSON and then to bytes
        json_data = json.dumps(data)
        bytes_data = bytes(json_data, "utf-8")

        # Create a padder instance and pad the data
        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(bytes_data) + padder.finalize()

        # Generate a random IV and create a cipher object
        iv = os.urandom(16)
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
        encryptor = cipher.encryptor()

        # Encrypt the padded data
        encrypted_data = encryptor.update(padded_data) + encryptor.finalize()
        b64_encrypted_data = base64.b64encode(encrypted_data).decode("utf-8")
        b64_iv = base64.b64encode(iv).decode("utf-8")
        encrypted_data_obj = {"encrypted_data": b64_encrypted_data, "iv": b64_iv}

        return json.dumps(encrypted_data_obj).encode("utf-8")

    @staticmethod
    def decrypt(data: bytes, key: bytes) -> Union[int, str, list, dict]:
        # Load the encrypted data object
        data_obj = json.loads(data.decode("utf-8"))
        encrypted_data = base64.b64decode(data_obj.get("encrypted_data"))
        iv = base64.b64decode(data_obj.get("iv"))

        # Create a cipher object
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
        decryptor = cipher.decryptor()

        # Decrypt the data
        decrypted_data = decryptor.update(encrypted_data) + decryptor.finalize()
        unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
        unpadded_data = unpadder.update(decrypted_data) + unpadder.finalize()

        return json.loads(unpadded_data.decode("utf-8"))



import pytest

from crypto_vault.crypto_vault import Encryption


def test_generate_key():
    """
    Test 256 bits encryption key generation, ensure size
    """
    key = Encryption.generate_key()
    assert len(key) == 32


def test_unique_key():
    """
    Test 256 bits encryption key generation, ensure uniqueness
    """
    key1 = Encryption.generate_key()
    key2 = Encryption.generate_key()
    assert key1 != key2


@pytest.mark.parametrize("data", ["Hello world!", 42, [1, 2, 3], {"key": "value"}])
def test_encrypt_decrypt(data):
    """
    Test encryption and decryption
    """
    key = Encryption.generate_key()
    encrypted_data = Encryption.encrypt(data, key)
    decrypted_data = Encryption.decrypt(encrypted_data, key)
    assert data == decrypted_data


def test_cant_decrypt_with_wrong_key():
    """
    Test decryption with wrong key
    """
    key1 = Encryption.generate_key()
    key2 = Encryption.generate_key()
    data = "Hello world!"
    encrypted_data = Encryption.encrypt(data, key1)
    with pytest.raises(Exception):
        Encryption.decrypt(encrypted_data, key2)



from pydantic import BaseModel
from eth_typing import Address


class Settings(BaseModel):
    CONTRACT_ADDRESS: Address = "0x276c89d95300b11b8ACEAE24a2dbc4bc3BaB69F5"
    ABI_FILENAME: str = "abi.json"


settings = Settings()



from setuptools import setup, find_packages

setup(
    name='crypto-vault',
    version='0.1.0b1',
    packages=find_packages(exclude=('tests', 'tests.*')),
    package_data={'crypto_vault': ['abi.json']},
    include_package_data=True,
    description='Efficiently encrypt data off-chain and store securely on the blockchain',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Mikhail Antonov',
    author_email='allelementaryfor@gmail.com',
    url='https://github.com/allelementary/crypto-vault-off-chain',
    install_requires=[
        "cryptography==42.0.5",
        "pydantic==2.6.4",
        "web3==6.15.1",
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: OS Independent',
    ],
)



import json
import os
from typing import Optional

from eth_typing import Address
from web3 import Web3

from crypto_vault.config import settings
from crypto_vault.encryption import Encryption


class CryptoVault(Encryption):
    """
    Store, update and retrieve on-chain data,
    including encryption, decryption and sending blockchain transactions
    """
    contract_address: Address = settings.CONTRACT_ADDRESS

    def __init__(
        self, app: str, env: str, private_key: str, encryption_key: bytes, http_provider: str
    ) -> None:
        self.app = app
        self.env = env
        self.private_key = private_key
        self.encryption_key = encryption_key
        self.w3 = Web3(Web3.HTTPProvider(http_provider))

    @property
    def abi(self):
        filename = settings.ABI_FILENAME
        current_dir = os.path.dirname(__file__)
        path = os.path.join(current_dir, filename)
        with open(path, "r") as file:
            abi = file.read()
        return json.loads(abi).get("abi")

    @property
    def user_address(self):
        return self.w3.eth.account.from_key(self.private_key).address

    def store(self, data: dict):
        # Get and decrypt existing data
        contract = self.w3.eth.contract(address=self.contract_address, abi=self.abi)
        encrypted_data = self.encrypt(data=data, key=self.encryption_key)

        # Create a transaction
        tx = contract.functions.store(
            self.w3.to_hex(text=self.app),
            self.w3.to_hex(text=self.env),
            self.w3.to_hex(encrypted_data)
        ).build_transaction(
            {
                "from": self.user_address,
                "nonce": self.w3.eth.get_transaction_count(self.user_address),
                "gasPrice": self.w3.eth.gas_price,
            }
        )

        # Sign and send the transaction
        signed_tx = self.w3.eth.account.sign_transaction(tx, self.private_key)
        tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)

        return tx_hash.hex()

    def retrieve(self, value_name: Optional[str] = None) -> dict:
        """Retrieve decrypted data from the storage"""
        # Instantiate the contract
        contract = self.w3.eth.contract(address=self.contract_address, abi=self.abi)

        # Retrieve the encrypted data from the blockchain
        encrypted_data = contract.functions.retrieve(
            self.w3.to_hex(text=self.app), self.w3.to_hex(text=self.env)
        ).call({"from": self.user_address})

        # Decrypt the data
        decrypted_data = self.decrypt(data=encrypted_data, key=self.encryption_key)
        if value_name:
            return decrypted_data.get(value_name)
        return decrypted_data

    def update(self, data: dict):
        # Get and decrypt existing data
        contract = self.w3.eth.contract(address=self.contract_address, abi=self.abi)
        existing_data = contract.functions.retrieve(
            self.w3.to_hex(text=self.app), self.w3.to_hex(text=self.env)
        ).call(
            {"from": self.user_address}
        )
        if existing_data:
            existing_data = self.decrypt(data=existing_data, key=self.encryption_key)
        else:
            existing_data = {}

        # Update and encrypt the data
        existing_data.update(data)
        encrypted_data = self.encrypt(data=existing_data, key=self.encryption_key)

        # Prepare the transaction to store the combined data
        tx_data = {
            "from": self.user_address,
            "nonce": self.w3.eth.get_transaction_count(self.user_address),
            "gasPrice": self.w3.eth.gas_price,
        }
        tx = contract.functions.store(
            self.w3.to_hex(text=self.app), self.w3.to_hex(text=self.env), self.w3.to_hex(encrypted_data)
        ).build_transaction(tx_data)

        # Sign and send the transaction
        signed_tx = self.w3.eth.account.sign_transaction(tx, self.private_key)
        tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)

        return tx_hash.hex()
